{ Unit otte_math contains code that manages conversions between decimal and
  duodecimal systems. For technical reasons, the code cannot convert numbers
  longer than 18 decimal (or 16 duodecimal) digits.
------------------------------------------------------------------------------}

{$I otte_cfg.pas}
unit otte_math;

interface //===================================================================

type
  TotteBaseConv = function(const num: ansistring): ansistring;

function b10_to_b12(const num: ansistring): ansistring;
function b12_to_b10(const num: ansistring): ansistring;

implementation //==============================================================

const
  Max64Len   = 18; //int64 will deal with no more than 18 decimal digits
  Max64b12Len= 16; //int64 will deal with no more than 16 duodecimal digits
  chunk12len = 8;
  chunk12div = 429981696; //12^chunk12len;

//-----------------------------------------------------------------------------
function chunk12(a: cardinal): string[chunk12len];

const
  //array of digits. Note, that for digits '10' and '11' I used ':' and ';',
  //because they are after '9' in the ASCII, and that makes conversion to
  //duodecimal format easier.
  Digits: array[0..11] of char = '0123456789:;'; //NOT string[12] !

var
  i, rest: integer;

begin
  result := '00000000';
  i := chunk12len;

  while a > 0 do begin
    rest := a mod 12;
    a := a div 12;
    result[i] := digits[rest];
    dec(i);
  end;
end;

//-----------------------------------------------------------------------------
function b10_to_b12(const num: ansistring): ansistring;

var
  i, count: longint;
  zeroes: ansistring; //keeps leading zeroes if any
  NumVal: int64;
  rest: cardinal;

begin
  result := num;

  //move leading zeroes into safe place, if there are any
  count := 0;
  while (count < length(result)) and (result[count+1] = '0') do inc(count);
  if count > 0 then begin
    zeroes := Copy(result, 1, count);
    Delete(result, 1, count);
  end;

  //isn't the numeral too long?
  if length(result) > Max64Len then setlength(result, max64Len);

  //do the conversion
  if length(result) > chunk12len then begin //int64 is enough
    Val(result, NumVal, i);
    result := '';

    while NumVal <> 0 do begin
      rest := NumVal mod chunk12div;
      NumVal := NumVal div chunk12div;
      result := chunk12(rest) + result;
    end;
  end
  else if length(result) > 0 then begin //cardinal is enough
    Val(result, rest, i);
    result := chunk12(rest);
  end;

  //Chunk12 produces spare leading zeroes. They must be removed.
  count := 0;
  while (count < length(result)) and (result[count+1] = '0') do inc(count);
  if count > 0 then Delete(result, 1, count);

  //add leading zeroes, if there were any before the conversion.

  { NOTE: This is of course an error in duodecimal fractions :(
    I want to add "IsFraction" parameter in future versions,
    but I need Galadhorn's advice before this...
  }
  
  result := zeroes + result;
end;

//-----------------------------------------------------------------------------
function b12_to_b10(const num: ansistring): ansistring;

var
  i, count: longint;
  zeroes: ansistring; //keeps leading zeroes if any
  NumVal, base: int64;

begin
  result := num;
  
  //move leading tengwar zeroes into safe place, if there are any
  count := 0;
  while (count < length(result)) and (result[count+1] = #240) do inc(count);
  if count > 0 then begin
    zeroes := Copy(result, 1, count);
    Delete(result, 1, count);
  end;

  //isn't the numeral too long?
  if length(result) > Max64b12Len then setlength(result, max64b12Len);

  //do the conversion
  if length(result) > 0 then begin
    base := 1;
    NumVal := 0;

    for i := length(result) downto 1 do begin
      NumVal := NumVal + base * (byte(result[i]) - 240);
      if i > 1 then base := base * 12;
    end;

    str(NumVal, result);
  end;
  
  //back to tengwar digits!
  for i := 1 to length(result) do inc(result[i], 240-byte('0'));

  //add leading zeroes, if there were any before the conversion.
  result := zeroes + result;
end;

end.