program teng;

{
    TENG v. 0.9.5, 2005
    Michal Nowakowski, <miszka@limes.com.pl>
    
    A fast, command-line transcriber (not translator!) from Latin
    alphabet to Tolkien's Tengwar and back. Compatible with Tengwar
    Scribe *.mod files. For more info, see the manual.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.
}

{$I otte/otte_cfg.pas}

uses
  cmem,
  unix,
  getopts,
  tengMessages,
  otte in 'otte/otte.pas',
  otte_ctype in 'otte/otte_ctype.pas';

const
  Slash = '/';

  Arguments = '?hum:i:o:l:trxcdbqB';
  Verbose:    boolean = true;
  BackTran:   boolean = false;
  NonTengwar: boolean = false;{}
  InputFile:  AnsiString = '';
  OutputFile: AnsiString = '';
  ModeFile:   AnsiString = '';
  Locale:     AnsiString = '';
  MsgFile:    AnsiString = 'teng.msg.'{locale.txt};
  LineNUm:    Longint    = 0;

  longind: longint = 0;
  longopts: array [0..16] of TOption = (
    (name:'version';     has_arg:0; flag:nil; value:'V'),
    (name:'usage';       has_arg:0; flag:nil; value:'u'),
    (name:'help';        has_arg:0; flag:nil; value:'h'),
    (name:'quiet';       has_arg:0; flag:nil; value:'q'),
    (name:'list-modes';  has_arg:0; flag:nil; value:'M'),
    (name:'locale';      has_arg:1; flag:nil; value:'l'),
    (name:'mode';        has_arg:1; flag:nil; value:'m'),
    (name:'input';       has_arg:1; flag:nil; value:'i'),
    (name:'output';      has_arg:1; flag:nil; value:'o'),
    (name:'text-nums';   has_arg:0; flag:nil; value:'t'),
    (name:'revert-nums'; has_arg:0; flag:nil; value:'r'),
    (name:'duodecimal';  has_arg:0; flag:nil; value:'x'),
    (name:'circle-nums'; has_arg:0; flag:nil; value:'c'),
    (name:'dot-nums';    has_arg:0; flag:nil; value:'d'),
    (name:'bar-nums';    has_arg:0; flag:nil; value:'b'),
    (name:'back';        has_arg:0; flag:nil; value:'B'),
    (name:'';            has_arg:0; flag:nil; value:'0')
  );

var
  CurrentArg: char;
  otteResult: integer;
  engine: TotteTransEngine;
  s: AnsiString;
  i: integer;
  ExitProcNext: procedure;
  sec10, Csec10: word;

//-----------------------------------------------------------------------------
function isatty(fd: longint): longint; cdecl; external 'c' name 'isatty';
function clock: longint; cdecl; external 'c' name 'clock';

//-----------------------------------------------------------------------------
procedure AtExit;
begin
  close(input);
  close(output);

  if assigned(engine) then engine.destroy;
  if assigned(ExitProcNext) then ExitProcNext();
end;

//-----------------------------------------------------------------------------
function IsFilePath(const s: AnsiString): boolean;

var
  i: integer;

begin
  result := false;
  
  for i := 1 to length(s) do
  if s[i] = Slash then begin
    result := true;
    exit;
  end;
end;

//-----------------------------------------------------------------------------
function FileAvailable(const fname: AnsiString): boolean;

var f: file;

begin
  result := true;
  assign(f, fname);
  reset(f, 1);

  if IOResult<>0 then result := false
  else close(f);
end;

//-----------------------------------------------------------------------------
function FoundFile(var what: AnsiString; const where: array of AnsiString): Boolean;

var
  TmpPath: AnsiString;
  i: integer;
  
begin
  result := false;
  
  if IsFilePath(what) then begin
    what := fexpand(what);
    result := FileAvailable(what);
  end
  else begin
    for i := 0 to high(where) do begin
      TmpPath := where[i] + what;
      if FileAvailable(TmpPath) then begin
        what := TmpPath;
        result := true;
        exit;
      end;
    end;
    
    result := FileAvailable(fexpand(what));
  end;
end;

//-----------------------------------------------------------------------------
begin
  //Try to load messages file from various locations
  MsgFile := MsgFile + getlocale;
  if FoundFile(MsgFile, [fexpand('~/.teng/messages/'), '/usr/share/teng/messages/']) then
    LoadMessages(MsgFile);

  ExitProcNext := ExitProc;
  ExitProc := @AtExit;
  engine := TotteTransEngine.create;

  repeat
    CurrentArg := getlongopts(arguments, @longopts, longind);
  
    with engine do
    begin
      case CurrentArg of
      'B': if BackTran then Die(errParamDuplicated, 'B')
           else BackTran := true;
      't': if TxtAll then Die(errParamDuplicated, 't')
           else TxtAll := true;
      'r': if RevNum then Die(errParamDuplicated, 'r')
           else RevNum := true;
      'x': if B12Num then Die(errParamDuplicated, 'x')
           else B12NUm := true;
      'c': if CirNum then Die(errParamDuplicated, 'c')
           else CirNum := true;
      'b': if BarNum then Die(errParamDuplicated, 'b')
           else BarNum := true;
      'd': if DotNum then Die(errParamDuplicated, 'd')
           else DotNUm := true;
      'q': if not Verbose then Die(errParamDuplicated, 'v')
           else Verbose := false;
      'i': if OptArg = '' then Die(errEmptyFileName, 'i')
           else if InputFile <> '' then Die(errParamDuplicated, 'i')
           else InputFile := OptArg;
      'o': if OptArg = '' then Die(errEmptyFileName, 'o')
           else if OutputFile <> '' then Die(errParamDuplicated, 'o')
           else OutputFile := OptArg;
      'm': if OptArg = '' then Die(errEmptyFileName, 'm')
           else if ModeFile <> '' then Die(errParamDuplicated, 'm')
           else ModeFile := OptArg;
      'l': if OptArg = '' then Die(errEmptyFileName, 'l')
           else if Locale <> '' then Die(errParamDuplicated, 'l')
           else Locale := OptArg;
      'u': begin
             PrintUsage;
             halt(0);
           end;
      'V': begin
             PrintVersion;
	     halt(0);
	   end;
      'M': begin
	     FPexecLP('modlist', []);
	     die(errModlistNotRun, ''); //if execlp returned, something went wrong
	   end;	   
      else 
        if CurrentArg <> EndOfOptions then
          begin
             PrintHelp;
             halt(0);
          end;
      end;
    end;
  until CurrentArg = EndOfOptions;

  if ModeFile = '' then Die(errParamMandatory, 'm');
  
  if Verbose then writeln(stderr, ProgramVersion);
  
  //assign input file to stdin, if specified
  if InputFile <> '' then begin
    close(input);
    assign(input, InputFile);
    reset(input);
    if IOResult <> 0 then Die(errFileNotOpened, InputFile);
    if Verbose then
      writeln(stderr, ProgramMessages[msgOpening], ' ', InputFile, '...', ProgramMessages[msgDone]);
  end;
  
  //assign output file to stdout, if specified
  if OutputFile <> '' then begin
    close(output);
    assign(output, OutputFile);
    rewrite(output);
    if IOResult <> 0 then Die(errFileNotOpened, OutputFile);
    if Verbose then
      writeln(stderr, ProgramMessages[msgOpening], ' ', OutputFile, '...', ProgramMessages[msgDone]);
  end;
  
  //Try to init locale
  otteResult := engine.mode.CharTransf.SetLocale(locale);
  if otteResult <> 0 then Die(errOtte-otteResult, locale);
  
  //Try to find mode file in various locations
  if not FoundFile(ModeFile, [fexpand('~/.teng/modes/'), '/usr/share/teng/modes/']) then begin
    s := ModeFile;
    ModeFile := ModeFile + '.mod';
    
    if not FoundFile(ModeFile, [fexpand('~/.teng/modes/'), '/usr/share/teng/modes/']) then
      Die(errFileNotOpened, s);
  end;

  //Try to load mode
  otteResult := engine.mode.load(ModeFile);
  if otteResult <> 0 then Die(errOtte-otteResult, ModeFile);
  if Verbose then
    writeln(stderr, ProgramMessages[msgOpening], ' ', ModeFile, '...', ProgramMessages[msgDone]);
  
  //unset verbosity if interactive session
  //this will suppress transcribed line number display
  if ((isatty(0)=1) and (InputFile = '')) or
     ((isatty(1)=1) and (OutputFile = '')) then Verbose := false;

  //now the transcription begins
  engine.GetReady;

  if Verbose then begin
    write(stderr, ProgramMessages[msgTranscribing], '...1', #08);
    Csec10 := clock div 100000;
  end;

  while (ioresult = 0) and not eof do begin
    if Verbose then begin
      inc(linenum);
      sec10 := clock div 100000;

      if sec10 <> Csec10 then begin
        Csec10 := sec10;
        str(linenum, s);
        write(stderr, s);
        for i := 1 to length(s) do write(stderr, #08);
      end;
    end;

    readln(s);
    writeln(engine.TranscribeLine(BackTran, s));
  end;
  
  //its over
  if Verbose then begin
    str(linenum, s);
    for i := 1 to length(s) do write(stderr, ' ');
    for i := 1 to length(s) do write(stderr, #08);
    writeln(stderr, ProgramMessages[msgDone]);
  end;
  
  engine.destroy;
  Exitproc := @ExitProcNext;
  close(input);
  close(output);
end.
